/* Copyright (C) 2015-2021, Wazuh Inc.
 * Copyright (C) 2009 Trend Micro Inc.
 * All right reserved.
 *
 * This program is free software; you can redistribute it
 * and/or modify it under the terms of the GNU General Public
 * License (version 2) as published by the FSF - Free Software
 * Foundation
 */

#include "shared.h"
#include "os_net/os_net.h"
#ifndef WIN32
int get_agent_ip(char *agent_ip)
{
#if defined (__linux__) || defined (__MACH__) || defined (sun) || defined(FreeBSD) || defined(OpenBSD)
    int sock;
    int i,retcd=1;
    static const char * REQUEST = "host_ip";

    for (i = SOCK_ATTEMPTS; i > 0; --i) {
        if (sock = control_check_connection(), sock >= 0) {
            if (OS_SendUnix(sock, REQUEST, strlen(REQUEST)) < 0) {
                mdebug1("Error sending msg to control socket (%d) %s", errno, strerror(errno));
            }
            else{
                if (OS_RecvUnix(sock, IPSIZE, agent_ip) <= 0) {
                    mdebug1("Error receiving msg from control socket (%d) %s", errno, strerror(errno));
                    agent_ip[0] = '\0';		    
                }
            }

            close(sock);
            break;
        } else {
            mdebug2("Control module not yet available. Remaining attempts: %d", i - 1);
            sleep(1);
        }
    }

    if(sock < 0) {
	retcd=0;
        mdebug1("Cannot get the agent host's IP because the control module is not available: (%d) %s.", errno, strerror(errno));
    }
#endif
    return retcd;
}
#endif /* !WIN32 */
/* Handle syslog connections */
void handle(uid_t  uid)
{
    char buffer[OS_MAXSTR + 2];
    char srcip[IPSIZE + 1]= { '\0' };
    int udp_sock,m_queue;
    char *buffer_pt;
    ssize_t recv_b;
    struct sockaddr_in peer_info;
    socklen_t peer_size;
    /* Set the source IP */
    if(!get_agent_ip(srcip))
    {
	  merror_exit("fatal","can not get agent ip");
    }
    /* Using UDP. Fast, unreliable... perfect */
    udp_sock = OS_Bindportudp(514,"127.0.0.1",0);

    if (udp_sock < 0) {
            merror_exit(BIND_ERROR, 514, errno, strerror(errno));
    }


    /* Revoke privileges */
    if (Privsep_SetUser(uid) < 0) {
        merror_exit(SETUID_ERROR, USER, errno, strerror(errno));
    }

    /* Create PID */
    if (CreatePID(ARGV0, getpid()) < 0) {
        merror_exit(PID_ERROR);
    }
    /* Set peer size */
    peer_size = sizeof(peer_info);

    /* Initialize some variables */
    memset(buffer, '\0', OS_MAXSTR + 2);
    memset(&peer_info, 0, sizeof(struct sockaddr_in));

    /* Connect to the message queue infinitely */
    if ((m_queue = StartMQ(DEFAULTQUEUE, WRITE, INFINITE_OPENQ_ATTEMPTS)) < 0) {
        merror_exit(QUEUE_FATAL, DEFAULTQUEUE);
    }

    /* Infinite loop */
    while (1) {
	int retry;
        /* Receive message */
        recv_b = recvfrom(udp_sock, buffer, OS_MAXSTR, 0, (struct sockaddr *)&peer_info, &peer_size);

        /* Nothing received */
        if (recv_b <= 0) {
            continue;
        }

        /* Null-terminate the message */
        buffer[recv_b] = '\0';

        /* Remove newline */
        if (buffer[recv_b - 1] == '\n') {
            buffer[recv_b - 1] = '\0';
        }

	mdebug2("received:%s\n",buffer);

        buffer_pt=strchr(buffer,'>');
        if( buffer_pt) buffer_pt++;
	else buffer_pt=buffer;

        retry=0;
        while (SendMSG(m_queue, buffer_pt, srcip, SYSLOG_MQ) < 0) {
            merror(QUEUE_ERROR, DEFAULTQUEUE, strerror(errno));

            // Try to reconnect infinitely
            m_queue = StartMQ(DEFAULTQUEUE, WRITE, INFINITE_OPENQ_ATTEMPTS);

            minfo("Successfully reconnected to '%s'", DEFAULTQUEUE);

             retry++;
	     if(retry>10) break;
        }
    }
}


#ifndef ARGV0
#define ARGV0 "syslog-collector"
#endif


/* Prototypes */
static void help_syslog_collector(char *home_path) __attribute((noreturn));


/* Print help statement */
static void help_syslog_collector(char *home_path)
{
    print_header();
    print_out("  %s: -[Vhdtf] [-u user] [-g group] [-c config]", ARGV0);
    print_out("    -V          Version and license message");
    print_out("    -h          This help message");
    print_out("    -d          Execute in debug mode. This parameter");
    print_out("                can be specified multiple times");
    print_out("                to increase the debug level.");
    print_out("    -t          Test configuration");
    print_out("    -f          Run in foreground");
    print_out("    -u <user>   User to run as (default: %s)", USER);
    print_out("    -g <group>  Group to run as (default: %s)", GROUPGLOBAL);
    print_out("    -c <config> Configuration file to use (default: %s)", OSSECCONF);
    print_out(" ");
    os_free(home_path);
    exit(1);
}


int main(int argc, char **argv)
{
    int c = 0;
    int test_config = 0;
    int debug_level = 0;
    char *home_path = w_homedir(argv[0]);

    const char *user = USER;
    const char *group = GROUPGLOBAL;
    const char *cfg = OSSECCONF;

    uid_t uid;
    gid_t gid;

    int run_foreground = 0;

    /* Set the name */
    OS_SetName(ARGV0);

	/* Change working directory */
    if (chdir(home_path) == -1) {
        merror(CHDIR_ERROR, home_path, errno, strerror(errno));
        os_free(home_path);
        exit(1);
    }

    int syslog_collector_debug_level = getDefine_Int("syslog-collector", "debug", 0, 2);

    while ((c = getopt(argc, argv, "Vtdfhu:g:D:c:")) != -1) {
        switch (c) {
            case 'V':
                print_version();
                break;
            case 'h':
                help_syslog_collector(home_path);
                break;
            case 'd':
                nowDebug();
                debug_level = 1;
                break;
            case 'f':
                run_foreground = 1;
                break;
            case 'u':
                if (!optarg) {
                    merror_exit("-u needs an argument");
                }
                user = optarg;
                break;
            case 'g':
                if (!optarg) {
                    merror_exit("-g needs an argument");
                }
                group = optarg;
                break;
            case 't':
                test_config = 1;
                break;
            case 'D':
                if (!optarg) {
                    merror_exit("-D needs an argument");
                }
                mwarn("-D is deprecated.");
                break;
            default:
                help_syslog_collector(home_path);
                break;
        }
    }


    /* Check current debug_level
     * Command line setting takes precedence
     */
    if (debug_level == 0) {
        /* Get debug level */
        debug_level = syslog_collector_debug_level;
        while (debug_level != 0) {
            nowDebug();
            debug_level--;
        }
    }

    mdebug1(WAZUH_HOMEDIR, home_path);
    os_free(home_path);
    mdebug1(STARTUP_MSG, (int)getpid());

    /* Check if the user/group given are valid */
    uid = Privsep_GetUser(user);
    gid = Privsep_GetGroup(group);
    if (uid == (uid_t) - 1 || gid == (gid_t) - 1) {
        merror_exit(USER_ERROR, user, group, strerror(errno), errno);
    }

    /* Exit if test config */
    if (test_config) {
        exit(0);
    }

    /* Start the signal manipulation */
    StartSIG(ARGV0);

    handle(uid);
    

    return (0);
}
